/*******************************************************************************
 File:		lcd.c
 Compiler:	AVR-GCC 3.3
 Hardware:	AVR
 Date:		05.06.2009

 Basic routines for interfacing a HD44780U-based text LC-Display
 4bit mode, microcontroller I/O port, write only (rw pin -> GND),
 dual- or single-line display
*******************************************************************************/

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include "lcdmap.h"
#include "lcd.h"
#include <stdio.h>

// execution time for instructions (LCD specific)
#define EXETIMELONG  1640	// execution time (us) for CLR and HOME instruction
#define EXETIMESHORT   40	// execution time (us) for all other instructions



FILE lcd_str = FDEV_SETUP_STREAM (LCD_putc, NULL, _FDEV_SETUP_WRITE);


/*******************************************************************************
Low-level function to write byte to LCD controller
Input:	 b		byte to write to LCD
		 rs		1: data write
				0: instruction write
*******************************************************************************/
static void write(unsigned char b, _Bool rs)
{
	if (rs) PORT(RS_PORT) |= (1 << RS); else PORT(RS_PORT) &= ~(1 << RS);

	// output high nibble
	PORT(EN_PORT) |= (1 << EN);
	if (b & 0x80) PORT(DB7_PORT) |= (1 << DB7); else PORT(DB7_PORT) &= ~(1 << DB7);
	if (b & 0x40) PORT(DB6_PORT) |= (1 << DB6); else PORT(DB6_PORT) &= ~(1 << DB6);
	if (b & 0x20) PORT(DB5_PORT) |= (1 << DB5); else PORT(DB5_PORT) &= ~(1 << DB5);
	if (b & 0x10) PORT(DB4_PORT) |= (1 << DB4); else PORT(DB4_PORT) &= ~(1 << DB4);
	PORT(EN_PORT) &= ~(1 << EN);

	// output low nibble
	PORT(EN_PORT) |= (1 << EN);
	if (b & 0x08) PORT(DB7_PORT) |= (1 << DB7); else PORT(DB7_PORT) &= ~(1 << DB7);
	if (b & 0x04) PORT(DB6_PORT) |= (1 << DB6); else PORT(DB6_PORT) &= ~(1 << DB6);
	if (b & 0x02) PORT(DB5_PORT) |= (1 << DB5); else PORT(DB5_PORT) &= ~(1 << DB5);
	if (b & 0x01) PORT(DB4_PORT) |= (1 << DB4); else PORT(DB4_PORT) &= ~(1 << DB4);
	PORT(EN_PORT) &= ~(1 << EN);
}


/*******************************************************************************
Low-level function to write nibble to LCD controller
Input:	 byte, high nibble is written to LCD
*******************************************************************************/
static void write_nibble(unsigned char b)
{
	PORT(RS_PORT) &= ~(1 << RS);		// instruction

	// output high nibble
	PORT(EN_PORT) |= (1 << EN);
	if (b & 0x80) PORT(DB7_PORT) |= (1 << DB7); else PORT(DB7_PORT) &= ~(1 << DB7);
	if (b & 0x40) PORT(DB6_PORT) |= (1 << DB6); else PORT(DB6_PORT) &= ~(1 << DB6);
	if (b & 0x20) PORT(DB5_PORT) |= (1 << DB5); else PORT(DB5_PORT) &= ~(1 << DB5);
	if (b & 0x10) PORT(DB4_PORT) |= (1 << DB4); else PORT(DB4_PORT) &= ~(1 << DB4);
	PORT(EN_PORT) &= ~(1 << EN);
}


/*******************************************************************************
Initialize display, see HD44780 data sheet
After initialisation the display is off!
*******************************************************************************/
void LCD_init()
{
	// configure RW pin, if available
	#ifdef RW
	DDR(RW_PORT) |= (1 << RW);			// output
	PORT(RW_PORT) &= ~(1 << RW);		// write
	#endif

	// configure data pins as output
	DDR(DB7_PORT) |= (1 << DB7);
	DDR(DB6_PORT) |= (1 << DB6);
	DDR(DB5_PORT) |= (1 << DB5);
	DDR(DB4_PORT) |= (1 << DB4);

	// configure control pins as output
	DDR(RS_PORT) |= (1 << RS);
	DDR(EN_PORT) |= (1 << EN);

	_delay_us(15000);					// wait for more then 15ms

	write_nibble(0x30);
	_delay_us(4100);					// wait for more then 4.1ms

	write_nibble(0x30);
	_delay_us(100);						// wait for more then 100us

	write_nibble(0x30);
	_delay_us(EXETIMESHORT);

	write_nibble(0x20);					// Function Set: 4bit Mode
	_delay_us(EXETIMESHORT);

	#if DISPLAY_LINES == 1
	LCD_instruction(0x20);				// Function Set: 1 line, 5x8 dot character font
	#else
	LCD_instruction(0x28);				// Function Set: 2 lines, 5x8 dot character font
	#endif

	//LCD_instruction(LCD_DISP_OFF);		// Display off, Cursor off, Blinking off

	LCD_instruction(LCD_CLR);			// Clear display

	LCD_instruction(LCD_ENTRY_INC);	    // Entry Mode Set: Increment, no shift

	//LCD_instruction(LCD_DISP_ON);       // Display Einschalten
}


/*******************************************************************************
Load custom character. This function change the cursor position!
Input:	 pointer to array of char
		 character code (0x00 - 0x07)
*******************************************************************************/
void LCD_loadCustomChar(unsigned char* customCharArray, unsigned char charCode)
{
	register unsigned char i;

	charCode = (charCode << 3);	// calculate CG RAM address

	// copy the 8 bytes into CG (character generator) RAM
	for (i=0; i<8; i++)
	{
		write(0x40 | (charCode + i), 0);				// set CG RAM address
		_delay_us(EXETIMESHORT);

		write(pgm_read_byte(customCharArray + i), 1);	// write character data
		_delay_us(EXETIMESHORT);
	}
}


/*******************************************************************************
Send instruction to LCD unit
Input:	 Instruction code (use predefined codes from lcd.h)
*******************************************************************************/
void LCD_instruction(unsigned char cmd)
{
	write(cmd, 0);
	if (cmd==LCD_CLR || cmd==LCD_HOME)
		_delay_us(EXETIMELONG);
	else
		_delay_us(EXETIMESHORT);
}


/*******************************************************************************
Set cursor to specified position
Input:	 vertical position    (0: first line)
		 horizontal position  (0: left most position)
*******************************************************************************/
void LCD_setpos(unsigned char y, unsigned char x)
{
	if (y==0)							// first line
		write(0x80 | x, 0);
	else								// second line
		write(0x80 | (0x40+x), 0);

	_delay_us(EXETIMESHORT);
}


/*******************************************************************************
Display character at current cursor position
Function is used by stream object lcd_str
Input:	 character to be displayed
		 Pointer to FILE Object (shoud be NULL because Pointer is never used)
*******************************************************************************/
int LCD_putc(char c, FILE *stream)
{
	write(c, 1);
	_delay_us(EXETIMESHORT);
	return 0;
}


/*******************************************************************************
Display string without auto linefeed
Input:	 string to be displayed
*******************************************************************************/
void LCD_puts(const char *string)
{
	char c;

	while (c = *string++)
		LCD_putchar(c);

}
